# Refer to《Coherent Ising machine based on degenerate optical parametric oscillators》
# 《A fully programmable 100-spin coherent Ising machine with all-to-all connections》
# 《A coherent Ising machine for 2000-node optimization problems》
import numpy as np
from .. import solver


def RK45c_number(device, setup):

    # Physical layer input
    r_s = device.r_s
    r_p = device.r_p
    k = device.k
    As = np.sqrt(r_s * r_p / (2 * k**2))

    # Application layer input
    intensity = setup.intensity
    round_number = setup.round_number
    J = setup.couple_matrix
    N = J.shape[0]

    t_end = round_number

    sol_info = solver.RK45(device, setup, t_end)

    return sol_info



import numpy as np

from pycim import sampler
from .model.meanField import RK45meanField
from .model.c_number import RK45c_number
from .model.discrete import RK45Discrete
def singleSimulation(  device , setup , model = "discrete" , solver = "RK45" ):
    if( model == "discrete"):
        if(solver == "RK45"):
            # What is returned here are the in-phase component c and the gain np. sqrt (gain)
            sol_info = RK45Discrete(device , setup)
            return sol_info
        
    if( model == "c-number"):
        if(solver == "RK45"):
            # What is returned here is the in-phase component c
            sol_info = RK45c_number(device , setup)
            return sol_info
        
    if( model == "meanField"):
        if(solver == "RK45"):
            # What is returned here is the in-phase component c
            sol_info = RK45meanField(device , setup)
            return sol_info


# Run multiple times and return a list of cut values
def multiSimulation(  device , setup , step = 100,model = "discrete" , solver = "RK45" , ):

    if( model == "discrete"):
        if(solver == "RK45"):
            cut_list = np.zeros(step,)
            for i in range(step):
                sol_info = singleSimulation(device , setup , model = "discrete" , solver = "RK45" )
                cut = sampler.getCutValue(sol_info[0] , setup)
                cut_list[i] = cut
        return cut_list
    
    if( model == "c-number"):
        if(solver == "RK45"):
            cut_list = np.zeros(step,)
            for i in range(step):
                sol_info = singleSimulation(device , setup , model = "c-number" , solver = "RK45" )
                cut = sampler.getCutValue(sol_info.y , setup)
                cut_list[i] = cut
        return cut_list
    
    if( model == "meanField"):
        if(solver == "RK45"):
            cut_list = np.zeros(step,)
            for i in range(step):
                sol_info = singleSimulation(device , setup , model = "meanField" , solver = "RK45" )
                cut = sampler.getCutValue(sol_info.y , setup)
                cut_list[i] = cut
        return cut_list


# -*- coding: utf-8 -*-
"""
module: pycim

Function: CIM physical simulation and application solution verification

"""
# author: Peixiang Li, peixiangli@quanta.org.cn

from pycim import simulation
from pycim import utils
from pycim import competitor
from pycim import sampler
from .analyzer import *


__all__ = ["simulation", "utils", "competitor", "sampler"]



# -*- coding: utf-8 -*-
"""
module: simulation

Function: CIM Physical Simulation

"""
# author: Peixiang Li, peixiangli@quanta.org.cn


from .setup import setup
from .device import device
from .solver import RK4,RK45,eulerMethod
from .simulate import singleSimulation,multiSimulation

__all__ = ["setup","device",'const']
__all__ += ["print_info"]
__all__ += ["RK4","RK45",'eulerMethod']
__all__ += ["singleSimulation","multiSimulation"]


## Refer to《Traveling-wave model of coherent Ising machine based on fiber loop with
## pulse-pumped phase-sensitive amplifier》

import numpy as np
from .. import solver


def RK45Discrete(device, setup):

    # Physical layer input
    L = device.L_ppln
    eta = device.loss
    kappa = device.kappa
    tao = device.tao
    t_end = 1 * tao  # Travel time in PSA

    # Application layer input
    J = setup.couple_matrix
    N = J.shape[0]
    lunshu = setup.round_number
    intensity = setup.intensity
    init_Ep = setup.pump_schedule

    # Return parameters
    gain = np.zeros(
        (N, lunshu)
    )  # Used to represent the gain obtained throughout the entire process of a round (including losses, coupling, and noise effects)响）
    c = np.zeros((N, lunshu))  # The in-phase component, Es is c

    # Temporary parameters
    sign_value = np.zeros((N, lunshu - 1))  # Sign value of in-phase component
    sqrt_G_I = np.zeros((N, lunshu))  # Gain of in-phase component
    N_I = np.zeros((N, lunshu))  # Noise of in-phase component
    noise_size = 2e-6  # The magnitude of noise level

    # The default signal light phase here is 0/π
    N_I[:N, 0] = noise_size * np.random.normal(
        0, 0.5, N
    )  # The initial value comes from vacuum fluctuations
    u = np.zeros(
        2 * N,
    )  # 0~N-1 are Es, and the last N are Ep
    c[0:N, 0] = N_I[0:N, 0]

    for k in range(0, lunshu - 1):
        u[0:N] = c[0:N, k]
        u[N : 2 * N] = init_Ep[k]
        sol_info = solver.RK45(u, t_end, kappa)
        if sol_info.success == False:
            print("False!")
        Es = sol_info.y[0:N]

        ################## Debug processing after the occurrence of gain 0 in the discrete model####################
        for i in range(len(Es[:N, -1])):
            if Es[i, -1] == 0:
                tmp_sign = np.sign(Es[i, 0])
                Es[i, -1] = tmp_sign * noise_size * abs(np.random.normal(0, 0.5, 1))
        ##################                                          ####################

        sqrt_G_I[:N, k] = (
            Es[:N, -1] / Es[:N, 0]
        )  # The gain obtained by PSA for the in-phase component in this round

        ################## Debug processing after the occurrence of gain 0 in the discrete model####################
        for i in range(len(sqrt_G_I[:N, k])):
            if sqrt_G_I[i, k] == 0:
                sqrt_G_I[i, k] = sqrt_G_I[i, k - 1]
        ##################                                          ####################

        N_I[:N, k] = (
            noise_size
            * np.random.normal(0, 1, N)
            * np.sqrt(0.25 * (2 - eta) * (sqrt_G_I[:N, k] ** 2))
        )
        c[:N, k + 1] = (
            sqrt_G_I[:N, k] * np.sqrt(eta) * c[:N, k]
            + sqrt_G_I[:N, k] * intensity[k] * np.dot(c[:, k], J)
            + N_I[0:N, k]
        )
        gain[:N, k] = c[:N, k + 1] / c[:N, k]
    return c, sqrt_G_I


def RK4Discrete(device, setup):

    return


def eulerDiscrete(device, setup):

    return



from matplotlib import pyplot as plt
import numpy as np
from functools import singledispatch
import scipy
from .utils import Ising


# Reverse search aims to find the time point at which
# max-cut_malue is reached in steady state (discrete model)
def findSteadyTime(c, setup):

    x = np.linspace(0, setup.round_number, setup.round_number + 1)
    sign_value = np.sign(c[:, :])
    cut_value = -0.5 * Ising(setup.couple_matrix, sign_value) - 0.25 * np.sum(
        setup.couple_matrix
    )
    for ti in range(len(cut_value) - 1, -1, -1):
        if cut_value[ti] < max(cut_value):
            return x[ti + 1]


# Return the minimum Ising energy (discrete model)
@singledispatch
def getMinEnergy(sol_info: np.ndarray, setup):

    sign_value = np.sign(sol_info[:, :])
    ising_energy = Ising(setup.couple_matrix, sign_value)
    mim_IsingEnergy = min(ising_energy)
    return mim_IsingEnergy


# Return the minimum Ising energy (c-number model)
@getMinEnergy.register
def _(sol_info: scipy.integrate._ivp.ivp.OdeResult, setup):
    c = sol_info.y
    N = setup.couple_matrix.shape[0]
    sign_value = np.sign(c[:N, :])
    ising_energy = Ising(setup.couple_matrix, sign_value)
    mim_IsingEnergy = min(ising_energy)
    return mim_IsingEnergy


# The first time to find the minimum Ising energy (discrete model)
@singledispatch
def getMinEnergyTime(sol_info: np.ndarray, setup):

    sign_value = np.sign(sol_info[:, :])
    ising_energy = Ising(setup.couple_matrix, sign_value)
    mim_IsingEnergy_time = np.argmin(ising_energy)
    return mim_IsingEnergy_time


# The first time to find the minimum Ising energy (c-number model)
@getMinEnergyTime.register
def _(sol_info: scipy.integrate._ivp.ivp.OdeResult, setup):
    c = sol_info.y
    x = sol_info.t
    N = setup.couple_matrix.shape[0]
    sign_value = np.sign(c[:N, :])
    ising_energy = Ising(setup.couple_matrix, sign_value)
    mim_IsingEnergy_time = np.argmin(ising_energy)
    return x[mim_IsingEnergy_time]


# Find the bifurcation time point in this model, assuming that the saturation
# and bifurcation time points are the same (discrete model)
# Note: This method is only available under gradual pump settings,
# and other schemes may obtain incorrect values
@singledispatch
def findSaturationTime(Gain: np.ndarray, setup):
    Gain = Gain[0]
    x = np.linspace(0, setup.round_number, setup.round_number + 1)
    for ti in range(1, len(x) - 1):
        if Gain[ti] - Gain[ti - 1] < 0:
            return int(x[ti])


# Find the bifurcation time point in this model, assuming that the saturation
# and bifurcation time points are the same (c-number model)
# Note: This method is only available under gradual pump settings,
# and other schemes may obtain incorrect values
@findSaturationTime.register
def _(sol_info: scipy.integrate._ivp.ivp.OdeResult, setup):
    sol = sol_info.y
    x = sol_info.t
    for ti in range(0, len(x)):
        if abs(sol[0][ti]) > abs((0.1 * sol[0][-1])):
            return x[ti]


# OPO amplitude at saturation (discrete model)
def getSaturationAmplitude(Gain, sol_info, setup):

    SaturationTime = findSaturationTime(Gain, setup)
    SaturationAmplitude = abs(sol_info[0][int(SaturationTime)])
    return SaturationAmplitude


# Pump power at saturation (discrete model)
def getSaturationPower(Gain, setup):

    SaturationTime = findSaturationTime(Gain, setup)
    SaturationPower = setup.pump_schedule[int(SaturationTime)]
    return SaturationPower


# Draw the evolution diagram of OPO in-phase component (discrete model)
@singledispatch
def inPhase_graph(sol_info: np.ndarray, setup):

    N = len(sol_info[0:])
    t = np.linspace(0, setup.round_number - 1, setup.round_number)
    print(f"t.shape:{t.shape}")
    print(f"sol_info.shape:{sol_info.shape}")
    for i in range(N):
        plt.plot(t, sol_info[i, :], label=f"{i+1}")
    plt.xlabel('round number')
    plt.ylabel('in-phase amplitude')
    plt.show()


# Draw the evolution diagram of OPO in-phase component (c-number model)
@inPhase_graph.register
def _(sol_info: scipy.integrate._ivp.ivp.OdeResult, setup):

    c = sol_info.y
    x = sol_info.t
    N = setup.couple_matrix.shape[0]
    a_sat = abs(c[:N, -1])  # 饱和时的振幅,用作归一化
    N = len(setup.couple_matrix[0])
    for i in range(0, N):
        plt.plot(x, c[i, :] / a_sat[i], label=f"{i+1}")
    plt.xlabel('round  number')
    plt.ylabel('in-phase amplitude')
    plt.show()


# Draw the evolution diagram of Ising energy (discrete model)
@singledispatch
def energy_graph(sol_info: np.ndarray, setup):

    t = np.linspace(0, setup.round_number - 1, setup.round_number)
    x = t[:]
    sign_value = np.sign(sol_info[:, :])
    ising_energy = Ising(setup.couple_matrix, sign_value)
    plt.plot(x, ising_energy)
    plt.xlabel('round number')
    plt.ylabel('Ising energy')
    plt.show()


# Draw the evolution diagram of Ising energy (c-number model)
@energy_graph.register
def _(sol_info: scipy.integrate._ivp.ivp.OdeResult, setup):

    x = sol_info.t
    c = sol_info.y
    N = setup.couple_matrix.shape[0]
    sign_value = np.sign(c[:N, :])
    ising_energy = Ising(setup.couple_matrix, sign_value)
    plt.plot(x, ising_energy)
    plt.xlabel('round number')
    plt.ylabel('Ising energy')
    plt.show()


# Draw the evolution diagram of gain (discrete model)
def gain_graph(gain, setup):
    sqrt_G_I = gain
    t = np.linspace(0, setup.round_number - 1, setup.round_number)
    plt.plot(t[:-1], 10 * np.log10((sqrt_G_I[0][:-1] ** 2)))
    plt.xlabel('round number')
    plt.ylabel('gain(dB)')
    plt.show()
    return



import numpy as np
from ..utils import Ising


def bias_for_aux_spin(couple_mat, h_vector, N_aux=1):
    assert (
        len(h_vector) == couple_mat.shape[0]
    ), "h_vector length must match the number of rows in couple_mat"
    N = len(h_vector)
    H_mat = np.zeros((N, N_aux))
    H_mat[:, :] = h_vector[:, np.newaxis] / N_aux
    top = np.hstack([couple_mat, H_mat])
    F_mat = np.ones((N_aux, N_aux)) - np.eye(N_aux)
    bottom = np.hstack([H_mat.T, F_mat])
    combined_matrix = np.vstack([top, bottom])
    return combined_matrix


def get_offset(N_aux):
    if N_aux == 1:
        return -1
    elif N_aux > 1:
        return -(N_aux - 1) * N_aux / 2


def min_energy_for_aux_spin(sol_info: np.ndarray, setup, N_aux):
    c = sol_info.y
    N = setup.couple_matrix.shape[0]
    sign_value = np.sign(c[:N, :])
    ising_energy = Ising(setup.couple_matrix, sign_value)
    mim_IsingEnergy = min(ising_energy)
    return mim_IsingEnergy - get_offset(N_aux)


def bias_for_non_aux_spin(couple_mat, h_vector):
    assert (
        len(h_vector) == couple_mat.shape[0]
    ), "h_vector length must match the number of rows in couple_mat"
    np.fill_diagonal(couple_mat, h_vector)
    return couple_mat



from .. import solver


def RK45meanField(phyInput, applInput):

    # Physical layer input
    # PPLN length： m
    L_ppln = phyInput.L_ppln
    # Pump light wavelength： m
    lambda_in = phyInput.lambda_in

    # Application layer input
    intensity = applInput.intensity
    round_number = applInput.round_number
    J = applInput.couple_matrix

    N = J.shape[0]
    t_end = round_number

    sol_info = solver.RK45(t_end, phyInput, applInput)

    return sol_info



##########Define some constants used here
import numpy as np
global c 
global deff
global eps0
c = 2.99792458e8  # light speed m/s
deff = 28e-12  # Quadratic nonlinear coefficient 单位m/V
eps0 = 8.854187817e-12  # Vacuum dielectric constant F/m
h_bar = 1.05457266e-34  # Reduced Planck constant J·s

# Variable parameters, if they are variables, they can be deleted
global n1
global n2
n1 = 2  # Refractive index
n2 = 2  # Refractive index

# Define coefficient matrix
def matrix_J(N):
    J=np.ones((N,N))
    for i in range(0,N):
        J[i][i]=0   
    return J  


import random
import numpy as np


def e_J(J):
    return np.sqrt(np.sum(J * J))


def esing(J, s):
    col, row = J.shape
    e = 0
    for c in range(col):
        for r in range(row):
            e = e - 0.5 * J[c][r] * s[0][c] * s[0][r]
    return e


# P (i) is a (t), which increases from 0 to 1 after going through steps
def p(s: int):
    return s * 0.005


def dsb(J):
    size = J.shape[0]
    det_t = 0.5  # time step
    step = 200  # Evolution steps of KPO
    det = 1  # a0 Positive constant
    ksi0 = (
        0.5 * np.sqrt(size - 1) / e_J(J)
    )  # c0 Positive constant defined based on random matrix theory
    # a0 c0 all can be parameters to be optimized.
    x = np.zeros([1, size], dtype=np.float64)
    y = np.array(np.random.uniform(-0.1, 0.1, size=(1, x.size)), dtype=np.float64)
    for i in range(step + 1):
        y = y + det_t * (ksi0 * np.dot(np.sign(x), J) - (det - p(i)) * x)
        x = x + det * det_t * y
        y = np.where(np.abs(x) > 1, 0, y)
        x = np.where(np.abs(x) > 1, np.sign(x), x)
    return np.sign(x)


# Simulated Annealing Algorithm
## Refer to 《Massively Simulating Adiabatic Bifurcations with FPGA to Solve Combinatorial Optimization》
def SA(J):

    size = J.shape[0]
    T = 1000  # initial temperature
    r = 0.99  # step
    T_min = 0.001  # Final temperature
    s = np.ones([1, size], dtype=np.int32)  # Initial state all 1, initial spin all 1
    i = 0
    while T > T_min:
        select_num = random.randint(
            0, size - 1
        )  # Randomly select an integer between 0 and size-1
        dE = (
            2 * s[0][select_num] * np.sum(J[select_num] * s[0])
        )  # △E The energy change caused by flipping a spin
        if np.exp(-(dE / T)) > random.uniform(
            0, 1
        ):  # Add noise based on temperature to break out of the local minimum Metropolis rule: accept new states with probability
            # Flip s
            s[0][select_num] = -s[0][select_num]
        else:
            pass
        T = r * T
        i += 1
    cut_value = -0.5 * esing(J, s) - 0.25 * np.sum(J)

    return cut_value


# Simulated bifurcation algorithm
## Refer to《High-performance combinatorial optimization based on classical mechanics》
def SB(J, step):

    i = 0
    max_cut_value_sb = 0
    while i < step:
        s = dsb(J)
        if (-0.5 * esing(J, s) - 0.25 * np.sum(J)) > max_cut_value_sb:
            max_cut_value_sb = -0.5 * esing(J, s) - 0.25 * np.sum(J)
        i += 1
    # print("s:",s)#The spin configuration of the solution is this matrix
    return max_cut_value_sb


# Fixed Greedy Algorithm with Polynomial Time Complexity
def SG(J):
    J = -J

    def weight_all(i, S):
        weight = 0
        for j in S:
            weight += J[i][j]
        return weight

    cut_value = float('-inf')
    N = len(J[0])
    V_ = np.linspace(0, N - 1, N)  # V`=V
    # Find the edge with the highest weight in J
    max_weight = J[0][1]
    row = 0
    col = 1
    for r in range(0, N - 1):
        for c in range(r + 1, N):
            if J[r][c] > max_weight:
                max_weight = J[r][c]
                row = r
                col = c
    cut_value = max_weight
    V_ = np.delete(V_, np.where(V_ == row))
    V_ = np.delete(V_, np.where(V_ == col))
    S1 = np.array([row])
    S2 = np.array([col])
    score = [float('-inf')] * N

    for j in range(1, N - 1):

        for i in V_:
            i = int(i)
            score[i] = max(weight_all(i, S1), weight_all(i, S2))
        i_ = score.index(max(score))

        if weight_all(i_, S1) > weight_all(i_, S2):
            S2 = np.append(S2, i_)
        else:
            S1 = np.append(S1, i_)

        V_ = np.delete(V_, np.where(V_ == i_))
        score[i_] = float("-inf")
        cut_value = cut_value + max(weight_all(i_, S1), weight_all(i_, S2))

    return cut_value


# Heuristic algorithm that can guarantee at least xx approximation ratio
def GW_SDP():
    return



import numpy as np

from pycim.utils import const

class device:
    def __init__(self):
        ##### Parameters required in the discrete (traveling wave) model
        # light speed m/s
        self.c  = 2.99792458e8  
        # Cavity length noraml: m
        self.L_cavity = 100 
        # Set the loss of the cavity, normal : dB
        self.loss = 10**(-11/10)
        # Set the gain coefficient of the PPLN crystals, normal: W^(-1/2)
        self.kappa = 130 * self.c 
        # Set the length of PPLN crystals, normal : m
        self.L_ppln = 0.05 
        # round trip time
        self.T_rt = self.L_cavity / self.c
        # Travel time in PPLN
        self.tao = self.L_ppln / self.c

        ##### Parameters required in the c-number model
        #Signal light, photon attenuation rate
        self.r_s = 1
        #Pump light, photon attenuation rate
        self.r_p = 10264
        # PPLN crystal gain coefficient
        self.k = 130 
        # Threshold pump amplitude
        self.F_th = (self.r_s*np.sqrt(self.r_p))/(4*self.k)
        self.As = np.sqrt(self.r_s*self.r_p/(2*self.k**2))
        self.g2 = 1/self.As

        ##### Parameters required in meanFiled model
        # Pump light wavelength 780nm
        self.lambda_in = 780e-9
        # Pump light frequency
        self.w1 = const.c / self.lambda_in
        # The coefficients of the coupled wave equation terms
        self.eps = 130  # equal to κ  
        # Refractive index    0.7179
        self.n = self.w1*2*const.deff/(self.eps * const.c)
        # Gain coefficient
        self.G0 = np.e**(2 * self.eps * self.L_ppln * np.sqrt(0.03796) ) ###自定义的  
        # Uncoupled threshold power
        self.b0 = np.log(self.G0)/(2 * self.eps * self.L_ppln)  
        # Saturation parameter
        self.beta = (self.G0-1-np.log(self.G0))/(4*(self.b0**2)*np.log(self.G0)) 



# Calculate Ising energy
import numpy as np


def Ising(J, sign_value):
    diag = np.diag(J)
    col, row = J.shape
    e = 0
    for c in range(col):
        for r in range(c + 1, row):
            e += -J[c][r] * sign_value[c] * sign_value[r]
    return e - np.dot(diag, sign_value)



# -*- coding: utf-8 -*-
"""
module: utils

Function: Some tools

"""
# author: Peixiang Li, peixiangli@quanta.org.cn

from .getIsingEnergy import Ising
from .file_J import read_J, tmp_read_J
from . import const
from .bias import bias_for_aux_spin, bias_for_non_aux_spin, min_energy_for_aux_spin

__all__ = [
    "Ising",
    "read_J",
    "tmp_read_J",
    "const",
    'bias_for_aux_spin',
    'bias_for_non_aux_spin',
    'min_energy_for_aux_spin',
]



import numpy as np


class setup:

    def __init__(self):

        def pump_schedule(t):
            return np.sqrt((1 / 760) * t)

        # Set the couple matrix
        # self.couple_matrix = - np.array([
        #         [0, 1, 0, 1, 1, 0, 0, 1, 1, 0],
        #         [1, 0, 1, 0, 0, 1, 1, 1, 0, 0],
        #         [0, 1, 0, 1, 1, 0, 0, 0, 1, 0],
        #         [1, 0, 1, 0, 0, 1, 1, 0 ,1, 0],
        #         [1, 0, 1, 0, 0, 1, 0, 1, 0, 1],
        #         [0, 1, 0, 1, 1, 0, 0, 0, 1, 1],
        #         [0, 1, 0, 1, 0, 0, 0, 0, 0, 1],
        #         [1, 1, 0, 0, 1, 0, 0, 0, 1, 0],
        #         [1, 0, 1, 1, 0, 1, 0, 1, 0, 1],
        #         [0, 0, 0, 0, 1, 1, 1, 0, 1, 0]])
        # The number of OPOs in the cavity
        self.N = 10
        # # Set the couple matrix
        self.couple_matrix = -np.zeros((self.N, self.N))
        # Round trip number
        self.round_number = 1500
        # Set the coupling intensity
        self.intensity = 0.03 * np.ones(
            self.round_number,
        )
        # Set the pump of the CIM system, normal:W
        self.pump_schedule = np.sqrt(
            5e-5 * np.linspace(1, self.round_number, self.round_number)
        )
        # Normalized pump power for c-number and meanFiled models
        self.p = pump_schedule

        self.N_aux = 0



from matplotlib import pyplot as plt
import numpy as np
from .utils import Ising
from mpl_toolkits.axes_grid1 import host_subplot
import mpl_toolkits.axisartist as AA
from functools import singledispatch
import scipy


def get_offset(N_aux):
    if N_aux == 1:
        return -1
    elif N_aux > 1:
        return -(N_aux - 1) * N_aux / 2


# The time to first find the optimal solution (discrete model)
@singledispatch
def getSolutionTime(sol_info: np.ndarray, setup):

    sign_value = np.sign(sol_info[:, :])
    ising_energy = Ising(setup.couple_matrix, sign_value)
    time_to_min_ising_energy = np.argmin(ising_energy)
    return time_to_min_ising_energy


# The time to first find the optimal solution (c-number model)
@getSolutionTime.register
def _(sol_info: scipy.integrate._ivp.ivp.OdeResult, setup):
    c = sol_info.y
    x = sol_info.t
    N = setup.couple_matrix.shape[0]
    sign_value = np.sign(c[:N, :])
    ising_energy = Ising(setup.couple_matrix, sign_value)
    time_to_min_ising_energy = np.argmin(ising_energy)
    return x[time_to_min_ising_energy]


# Classification configuration of optimal solution (discrete model)
@singledispatch
def getSolution(sol_info: np.ndarray, setup):

    sign_value = np.sign(sol_info[:, :])
    ising_energy = Ising(setup.couple_matrix, sign_value)
    time_to_min_ising_energy = np.argmin(ising_energy)
    opt_sol = sign_value[:, time_to_min_ising_energy]
    return opt_sol


# Classification configuration of optimal solution (c-number model)
# Note: The classification configuration obtained by the
# c-number model includes in-phase and quadrature components.
@getSolution.register
def _(sol_info: scipy.integrate._ivp.ivp.OdeResult, setup):
    c = sol_info.y
    N = setup.couple_matrix.shape[0]
    sign_value = np.sign(c[:N, :])
    if setup.N_aux > 0:
        ising_energy = Ising(setup.couple_matrix, sign_value) - get_offset(setup.N_aux)
    else:
        ising_energy = Ising(setup.couple_matrix, sign_value)
    time_to_min_ising_energy = np.argmin(ising_energy)
    opt_sol = sign_value[:, time_to_min_ising_energy]
    return opt_sol


# The accuracy of max_cut and based_cut obtained through simulation (discrete model)
@singledispatch
def getAccuracy(sol_info: np.ndarray, setup, based_cut):

    sign_value = np.sign(sol_info[:, :])
    cut_value = -0.5 * Ising(setup.couple_matrix, sign_value) - 0.25 * np.sum(
        setup.couple_matrix
    )
    max_cut = max(cut_value)
    accuracy = max_cut / based_cut
    return accuracy


# The accuracy of max_cut and based_cut obtained through simulation (c-number model)
@getAccuracy.register
def _(sol_info: scipy.integrate._ivp.ivp.OdeResult, setup, based_cut):
    c = sol_info.y
    N = setup.couple_matrix.shape[0]
    sign_value = np.sign(c[:N, :])
    cut_value = -0.5 * Ising(setup.couple_matrix, sign_value) - 0.25 * np.sum(
        setup.couple_matrix
    )
    max_cut = max(cut_value)
    accuracy = max_cut / based_cut
    return accuracy


# The maximum cut value of the optimal solution (discrete model)
@singledispatch
def getCutValue(sol_info: np.ndarray, setup):

    sign_value = np.sign(sol_info[:, :])

    cut_value = -0.5 * Ising(setup.couple_matrix, sign_value) - 0.25 * np.sum(
        setup.couple_matrix
    )
    max_cut = max(cut_value)
    return max_cut


# The maximum cut value of the optimal solution (c-number model)
@getCutValue.register
def _(sol_info: scipy.integrate._ivp.ivp.OdeResult, setup):
    c = sol_info.y
    N = setup.couple_matrix.shape[0]
    sign_value = np.sign(c[:N, :])
    cut_value = -0.5 * Ising(setup.couple_matrix, sign_value) - 0.25 * np.sum(
        setup.couple_matrix
    )
    max_cut = max(cut_value)
    return max_cut


# Draw the evolution diagram of cut-value values (discrete model)
@singledispatch
def cutvalue_graph(sol_info: np.ndarray, setup):

    t = np.linspace(0, setup.round_number - 1, setup.round_number)
    x = t[:-1]
    sign_value = np.sign(sol_info[:, : setup.round_number - 1])
    ising_energy = Ising(setup.couple_matrix, sign_value)
    cut_value = -0.5 * Ising(setup.couple_matrix, sign_value) - 0.25 * np.sum(
        setup.couple_matrix
    )
    y1 = cut_value
    y2 = ising_energy
    host = host_subplot(111, axes_class=AA.Axes)
    plt.subplots_adjust(right=0.9)
    par1 = host.twinx()
    offset = 100
    par1.axis["right"].toggle(all=True)
    host.set_xlabel("round number")
    host.set_ylabel("cut value")
    par1.set_ylabel("ising energ")
    p1 = host.plot(x, y1, color='blue')
    p2 = par1.plot(x, y2)
    par1.invert_yaxis()
    plt.show()


# Draw the evolution diagram of cut_ralue value (c-number model)
@cutvalue_graph.register
def _(sol_info: scipy.integrate._ivp.ivp.OdeResult, setup):
    c = sol_info.y
    x = sol_info.t
    sign_value = np.sign(c)
    ising_energy = Ising(setup.couple_matrix, sign_value)
    cut_value = -0.5 * Ising(setup.couple_matrix, sign_value) - 0.25 * np.sum(
        setup.couple_matrix
    )
    y1 = cut_value
    y2 = ising_energy
    host = host_subplot(111, axes_class=AA.Axes)
    plt.subplots_adjust(right=0.9)
    par1 = host.twinx()
    offset = 100
    par1.axis["right"].toggle(all=True)
    host.set_xlabel("round number")
    host.set_ylabel("cut value")
    par1.set_ylabel("ising energ")
    p1 = host.plot(x, y1, color='blue')
    p2 = par1.plot(x, y2)
    par1.invert_yaxis()
    plt.show()


# success rate
def getSuccessRate(cut_list, based_cut):

    p = sum(i >= based_cut for i in cut_list)
    SuccessRate = p / len(cut_list)
    return SuccessRate


# average cut
def getAveCutValue(cut_list, step):

    ave_cut = sum(cut_list) / step
    return ave_cut


# # Stable time to find the optimal solution
# def steadSolutionTime(c,setup):

#     sign_value = np.sign(c[:,:setup.round_number - 1])
#     cut_value = -0.5 * Ising(setup.couple_matrix, sign_value) - 0.25 * np.sum(setup.couple_matrix)
#     MaxCutValueTime = GainDownTime + np.argmax(cut_value[GainDownTime:]) # 稳定找到最大割值的时间点
#     return MaxCutValueTime



# -*- coding: utf-8 -*-
"""
module: model

Function: Provide three models of CIM

"""
# author: Peixiang Li, peixiangli@quanta.org.cn

from .c_number import RK45c_number
from .discrete import RK45Discrete,RK4Discrete,eulerDiscrete
from .meanField import RK45meanField
__all__ = ["RK45c_number","RK45Discrete","RK4Discrete","eulerDiscrete","RK45meanField"]


import numpy as np
from . import device
from scipy.integrate import solve_ivp
from functools import singledispatch


# discrete model
@singledispatch
def RK45(u: np.ndarray, t_end, kappa):
    def f(t, u):
        a = u
        N = int(len(a) / 2)
        b = np.zeros((2 * N,))
        # This equation is the coupled wave equation
        # with a default pump light phase of - π/2 and a signal light phase of 0/π
        b[:N] = kappa * a[N:] * a[:N]  # dEs/dz = kappa * Ap * Es
        b[N:] = -kappa * (a[:N] ** 2)  # dAp/dz = - kappa * Es²
        return b

    sol_info = solve_ivp(f, [0, t_end], u, method='RK45')
    return sol_info


# c-number model
@RK45.register
def _(device: device, setup, t_end):

    J = setup.couple_matrix
    # g2 = device.g2
    N = J.shape[0]
    pump_shcedule = setup.p
    g2 = 5e-3
    # noi_s = noise_size * np.random.randn(N)
    # noi_p = noise_size * np.random.randn(N)
    noi_s = g2 * np.random.choice((-1, 1), N)
    noi_p = g2 * np.random.choice((-1, 1), N)
    intensity = setup.intensity

    def f(t, u):
        a = u  # The first N are in-phase components, and the last N are quadrature components
        b = np.zeros_like(a)
        z = np.dot(a[0:N], J)
        b[:N] = (
            (pump_shcedule(t) - 1 - ((a[:N] ** 2 + a[N:] ** 2))) * a[:N]
            + intensity(t) * z
            + g2 * np.sqrt((a[:N] ** 2) + (a[N:] ** 2) + 0.5) * noi_s[:]
        )  # dcidt,
        b[N:] = (-pump_shcedule(t) - 1 - ((a[:N] ** 2 + a[N:] ** 2))) * a[
            N:
        ] + g2 * np.sqrt((a[:N] ** 2) + (a[N:] ** 2) + 0.5) * noi_p[
            :
        ]  # dsidt
        return b

    u = np.zeros((2 * N,))
    sol_info = solve_ivp(
        f,
        [0, t_end],
        u,
        method='RK45',
        first_step=1,
        t_eval=np.linspace(0, t_end, t_end + 1),
    )
    return sol_info


# meanField model
@RK45.register
def _(t_end: int, device, setup):

    J = setup.couple_matrix
    N = len(J[0])
    pump_shcedule = setup.p
    # g2 = device.g2
    # noi_s = 1e-7 * np.random.randn(N)

    g2 = 1e-5
    noi_s = g2 * np.random.choice((-1, 1), N)
    intensity = setup.intensity

    def f(t, u):
        a = u  # The first N are in-phase components, and the last N are quadrature components
        b = np.zeros_like(a)
        z = np.dot(a[0:N], J)
        b[:] = (pump_shcedule(t) - 1 - (a[:] ** 2)) * a[:] + intensity(t) * z
        return b

    u = np.array(np.random.uniform(-g2, g2, size=N), dtype=np.float64)
    sol_info = solve_ivp(
        f,
        [0, t_end],
        u,
        method='RK45',
        first_step=1,
        t_eval=np.linspace(0, t_end, t_end + 1),
    )
    return sol_info


def RK4():
    return


def eulerMethod():
    return



import numpy as np
def read_J(file_name):
    with open(file_name) as file:
        datas = file.readlines()
        line_1 = datas[0]
        size = int(line_1.split(" ")[0])
        J = np.zeros((size, size))
        for i in range(1, len(datas)):
            data = datas[i]
            edge = data.split(" ")
            start = int(edge[0])
            end = int(edge[1])
            # value = int(edge[2])
            value = edge[2]
            J[start - 1][end - 1] = value
            J[end - 1][start - 1] = value
        return -1 * J


#Read the file according to the J_ij setting in the paper
def tmp_read_J(file_name):
    with open(file_name) as file:
        datas = file.readlines()
        line_1 = datas[0]
        size = int(line_1.split(" ")[0])
        J = np.zeros((size, size))
        for i in range(1, len(datas)):
            data = datas[i]
            edge = data.split(" ")
            start = int(edge[0])
            end = int(edge[1])
            # value = int(edge[2])
            value = int(edge[2])
            if(value < 0):
                J[start - 1][end - 1] = 0
                J[end - 1][start - 1] = 0
            if(value > 0):
                J[start - 1][end - 1] = 1
                J[end - 1][start - 1] = 1
        return -1 * J
